/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Listing 63-9 */

  1 #include <sys/time.h>
  2 #include <sys/select.h>
  3 #include <fcntl.h>
  4 #include <signal.h>
  5 #include "tlpi_hdr.h"
  6
  7 static int pfd[2];                      /* File descriptors for pipe */
  8
  9 static void
 10 handler(int sig)
 11 {
 12     int savedErrno;                     /* In case we change 'errno' */
 13
 14     savedErrno = errno;
 15     if (write(pfd[1], "x", 1) == -1 && errno != EAGAIN)
 16         errExit("write");
 17     errno = savedErrno;
 18 }
 19
 20 int
 21 main(int argc, char *argv[])
 22 {
 23     fd_set readfds;
 24     int ready, nfds, flags;
 25     struct timeval timeout;
 26     struct timeval *pto;
 27     struct sigaction sa;
 28     char ch;
 29     int fd, j;
 30
 31     if (argc < 2 || strcmp(argv[1], "--help") == 0)
 32         usageErr("%s {timeout|-} fd...\n"
 33                 "\t\t('-' means infinite timeout)\n", argv[0]);
 34
 35     /* Initialize 'timeout', 'readfds', and 'nfds' for select() */
 36
 37     if (strcmp(argv[1], "-") == 0) {
 38         pto = NULL;                     /* Infinite timeout */
 39     } else {
 40         pto = &timeout;
 41         timeout.tv_sec = getLong(argv[1], 0, "timeout");
 42         timeout.tv_usec = 0;            /* No microseconds */
 43     }
 44
 45     nfds = 0;
 46
 47     /* Build the 'readfds' from the fd numbers given in command line */
 48
 49     FD_ZERO(&readfds);
 50     for (j = 2; j < argc; j++) {
 51         fd = getInt(argv[j], 0, "fd");
 52         if (fd >= FD_SETSIZE)
 53             cmdLineErr("file descriptor exceeds limit (%d)\n", FD_SETSIZE);
 54
 55         if (fd >= nfds)
 56             nfds = fd + 1;              /* Record maximum fd + 1 */
 57         FD_SET(fd, &readfds);
 58     }
 59
 60     if (pipe(pfd) == -1)
 61         errExit("pipe");
 62
 63     FD_SET(pfd[0], &readfds);           /* Add read end of pipe to 'readfds' */
 64     nfds = max(nfds, pfd[0] + 1);       /* And adjust 'nfds' if required */
 65
 66     flags = fcntl(pfd[0], F_GETFL);
 67     if (flags == -1)
 68         errExit("fcntl-F_GETFL");
 69     flags |= O_NONBLOCK;                /* Make read end nonblocking */
 70     if (fcntl(pfd[0], F_SETFL, flags) == -1)
 71         errExit("fcntl-F_SETFL");
 72
 73     flags = fcntl(pfd[1], F_GETFL);
 74     if (flags == -1)
 75         errExit("fcntl-F_GETFL");
 76     flags |= O_NONBLOCK;                /* Make write end nonblocking */
 77     if (fcntl(pfd[1], F_SETFL, flags) == -1)
 78         errExit("fcntl-F_SETFL");
 79
 80     sigemptyset(&sa.sa_mask);
 81     sa.sa_flags = SA_RESTART;           /* Restart interrupted reads()s */
 82     sa.sa_handler = handler;
 83     if (sigaction(SIGINT, &sa, NULL) == -1)
 84         errExit("sigaction");
 85
 86     while ((ready = select(nfds, &readfds, NULL, NULL, pto)) == -1 &&
 87             errno == EINTR)
 88         continue;                       /* Restart if interrupted by signal */
 89     if (ready == -1)                    /* Unexpected error */
 90         errExit("select");
 91
 92     if (FD_ISSET(pfd[0], &readfds)) {   /* Handler was called */
 93         printf("A signal was caught\n");
 94
 95         for (;;) {                      /* Consume bytes from pipe */
 96             if (read(pfd[0], &ch, 1) == -1) {
 97                 if (errno == EAGAIN)
 98                     break;              /* No more bytes */
 99                 else
100                     errExit("read");    /* Some other error */
101             }
102
103             /* Perform any actions that should be taken in response to signal */
104         }
105     }
106
107     /* Examine file descriptor sets returned by select() to see
108        which other file descriptors are ready */
109
110     printf("ready = %d\n", ready);
111     for (j = 2; j < argc; j++) {
112         fd = getInt(argv[j], 0, "fd");
113         printf("%d: %s\n", fd, FD_ISSET(fd, &readfds) ? "r" : "");
114     }
115
116     /* And check if read end of pipe is ready */
117
118     printf("%d: %s   (read end of pipe)\n", pfd[0],
119             FD_ISSET(pfd[0], &readfds) ? "r" : "");
120
121     if (pto != NULL)
122         printf("timeout after select(): %ld.%03ld\n",
123                (long) timeout.tv_sec, (long) timeout.tv_usec / 1000);
124
125     exit(EXIT_SUCCESS);
126 }
