/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Listing 37-3 */

  1 #include <sys/stat.h>
  2 #include <signal.h>
  3 #include "become_daemon.h"
  4 #include "tlpi_hdr.h"
  5
  6 static const char *LOG_FILE = "/tmp/ds.log";
  7 static const char *CONFIG_FILE = "/tmp/ds.conf";
  8
  9 #include <time.h>
 10 #include <stdarg.h>
 11
 12 static FILE *logfp;                 /* Log file stream */
 13
 14 /* Write a message to the log file. Handle variable length argument
 15    lists, with an initial format string (like printf(3), but without
 16    a trailing newline). Precede each message with a timestamp. */
 17
 18 static void
 19 logMessage(const char *format, ...)
 20 {
 21     va_list argList;
 22     const char *TIMESTAMP_FMT = "%F %X";        /* = YYYY-MM-DD HH:MM:SS */
 23 #define TS_BUF_SIZE sizeof("YYYY-MM-DD HH:MM:SS")       /* Includes '\0' */
 24     char timestamp[TS_BUF_SIZE];
 25     time_t t;
 26     struct tm *loc;
 27
 28     t = time(NULL);
 29     loc = localtime(&t);
 30     if (loc == NULL ||
 31            strftime(timestamp, TS_BUF_SIZE, TIMESTAMP_FMT, loc) == 0)
 32         fprintf(logfp, "???Unknown time????: ");
 33     else
 34         fprintf(logfp, "%s: ", timestamp);
 35
 36     va_start(argList, format);
 37     vfprintf(logfp, format, argList);
 38     fprintf(logfp, "\n");
 39     va_end(argList);
 40 }
 41
 42 /* Open the log file 'logFilename' */
 43
 44 static void
 45 logOpen(const char *logFilename)
 46 {
 47     mode_t m;
 48
 49     m = umask(077);
 50     logfp = fopen(logFilename, "a");
 51     umask(m);
 52
 53     /* If opening the log fails we can't display a message... */
 54
 55     if (logfp == NULL)
 56         exit(EXIT_FAILURE);
 57
 58     setbuf(logfp, NULL);                    /* Disable stdio buffering */
 59
 60     logMessage("Opened log file");
 61 }
 62
 63 /* Close the log file */
 64
 65 static void
 66 logClose(void)
 67 {
 68     logMessage("Closing log file");
 69     fclose(logfp);
 70 }
 71
 72 /* (Re)initialize from configuration file. In a real application
 73    we would of course have some daemon initialization parameters in
 74    this file. In this dummy version, we simply read a single line
 75    from the file and write it to the log. */
 76
 77 static void
 78 readConfigFile(const char *configFilename)
 79 {
 80     FILE *configfp;
 81 #define SBUF_SIZE 100
 82     char str[SBUF_SIZE];
 83
 84     configfp = fopen(configFilename, "r");
 85     if (configfp != NULL) {                 /* Ignore nonexistent file */
 86         if (fgets(str, SBUF_SIZE, configfp) == NULL)
 87             str[0] = '\0';
 88         else
 89             str[strlen(str) - 1] = '\0';    /* Strip trailing '\n' */
 90         logMessage("Read config file: %s", str);
 91         fclose(configfp);
 92     }
 93 }
 94
 95 static volatile sig_atomic_t hupReceived = 0;
 96                         /* Set nonzero on receipt of SIGHUP */
 97
 98 static void
 99 sighupHandler(int sig)
100 {
101     hupReceived = 1;
102 }
103
104 int
105 main(int argc, char *argv[])
106 {
107     const int SLEEP_TIME = 15;      /* Time to sleep between messages */
108     int count = 0;                  /* Number of completed SLEEP_TIME intervals */
109     int unslept;                    /* Time remaining in sleep interval */
110     struct sigaction sa;
111
112     sigemptyset(&sa.sa_mask);
113     sa.sa_flags = SA_RESTART;
114     sa.sa_handler = sighupHandler;
115     if (sigaction(SIGHUP, &sa, NULL) == -1)
116         errExit("sigaction");
117
118     if (becomeDaemon(0) == -1)
119         errExit("becomeDaemon");
120
121     logOpen(LOG_FILE);
122     readConfigFile(CONFIG_FILE);
123
124     unslept = SLEEP_TIME;
125
126     for (;;) {
127         unslept = sleep(unslept);       /* Returns > 0 if interrupted */
128
129         if (hupReceived) {              /* If we got SIGHUP... */
130             hupReceived = 0;            /* Get ready for next SIGHUP */
131             logClose();
132             logOpen(LOG_FILE);
133             readConfigFile(CONFIG_FILE);
134         }
135
136         if (unslept == 0) {             /* On completed interval */
137             count++;
138             logMessage("Main: %d", count);
139             unslept = SLEEP_TIME;       /* Reset interval */
140         }
141     }
142 }
