/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Supplementary program for Chapter 19 */

 1 #define _GNU_SOURCE             /* To get DN_* constants from <fcntl.h> */
 2 #include <fcntl.h>
 3 #include <signal.h>
 4 #include "tlpi_hdr.h"
 5
 6 static void             /* Print (optional) message and usage info, then exit */
 7 usageError(const char *progName, const char *msg)
 8 {
 9     if (msg != NULL)
10         fprintf(stderr, "%s", msg);
11
12     fprintf(stderr, "Usage: %s directory:[events]...\n", progName);
13     fprintf(stderr, "    Events are:\n"
14         "        a - access; A - attrib; c - create; d - delete\n"
15         "        m - modify; r - rename; M - multishot\n"
16         "        (default is all events)\n");
17     exit(EXIT_FAILURE);
18 }
19
20 static void
21 handler(int sig, siginfo_t *si, void *ucontext)
22 {
23     printf("Got event on descriptor %d\n", si->si_fd);
24                         /* UNSAFE (see Section 21.1.2) */
25 }
26
27 int
28 main(int argc, char *argv[])
29 {
30     struct sigaction sa;
31     int fd, events, fnum;
32     const int NOTIFY_SIG = SIGRTMIN;
33     char *p;
34
35     if (argc < 2 || strcmp(argv[1], "--help") == 0)
36         usageError(argv[0], NULL);
37
38     /* Establish handler for notification signal */
39
40     sa.sa_sigaction = handler;
41     sigemptyset(&sa.sa_mask);
42     sa.sa_flags = SA_SIGINFO;           /* So handler gets siginfo_t arg. */
43     if (sigaction(NOTIFY_SIG, &sa, NULL) == -1)
44         errExit("sigaction");
45
46     for (fnum = 1; fnum < argc; fnum++) {
47         p = strchr(argv[fnum], ':');    /* Look for optional ':' */
48
49         if (p == NULL) {                /* Default is all events + multishot */
50             events = DN_ACCESS | DN_ATTRIB | DN_CREATE | DN_DELETE |
51                      DN_MODIFY | DN_RENAME | DN_MULTISHOT;
52         } else {                        /* ':' present, parse event chars */
53             *p = '\0';                  /* Terminates directory component */
54             events = 0;
55             for (p++; *p != '\0'; p++) {
56                 switch (*p) {
57                 case 'a': events |= DN_ACCESS;          break;
58                 case 'A': events |= DN_ATTRIB;          break;
59                 case 'c': events |= DN_CREATE;          break;
60                 case 'd': events |= DN_DELETE;          break;
61                 case 'm': events |= DN_MODIFY;          break;
62                 case 'r': events |= DN_RENAME;          break;
63                 case 'M': events |= DN_MULTISHOT;       break;
64                 default:  usageError(argv[0], "Bad event character\n");
65                 }
66             }
67         }
68
69         /* Obtain a file descriptor for the directory to be monitored */
70
71         fd = open(argv[fnum], O_RDONLY);
72         if (fd == -1)
73             errExit("open");
74         printf("opened '%s' as file descriptor %d\n", argv[fnum], fd);
75
76         /* Use alternate signal instead of SIGIO for dnotify events */
77
78         if (fcntl(fd, F_SETSIG, NOTIFY_SIG) == -1)
79             errExit("fcntl - F_SETSIG");
80
81         /* Enable directory change notifications */
82
83         if (fcntl(fd, F_NOTIFY, events) == -1)
84             errExit("fcntl-F_NOTIFY");
85         printf("events: %o\n", (unsigned int) events);
86     }
87
88     for (;;)
89         pause();                        /* Wait for events */
90 }
