/*************************************************************************\
*                  Copyright (C) Michael Kerrisk, 2014.                   *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Affero General Public License as published   *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
* See the file COPYING.agpl-v3 for details.                               *
\*************************************************************************/

/* Listing 23-1 */

  1 #include <signal.h>
  2 #include <sys/time.h>
  3 #include <time.h>
  4 #include "tlpi_hdr.h"
  5
  6 static volatile sig_atomic_t gotAlarm = 0;
  7                         /* Set nonzero on receipt of SIGALRM */
  8
  9 /* Retrieve and display the real time, and (if 'includeTimer' is
 10    TRUE) the current value and interval for the ITIMER_REAL timer */
 11
 12 static void
 13 displayTimes(const char *msg, Boolean includeTimer)
 14 {
 15     struct itimerval itv;
 16     static struct timeval start;
 17     struct timeval curr;
 18     static int callNum = 0;             /* Number of calls to this function */
 19
 20     if (callNum == 0)                   /* Initialize elapsed time meter */
 21         if (gettimeofday(&start, NULL) == -1)
 22             errExit("gettimeofday");
 23
 24     if (callNum % 20 == 0)              /* Print header every 20 lines */
 25         printf("       Elapsed   Value Interval\n");
 26
 27     if (gettimeofday(&curr, NULL) == -1)
 28         errExit("gettimeofday");
 29     printf("%-7s %6.2f", msg, curr.tv_sec - start.tv_sec +
 30                         (curr.tv_usec - start.tv_usec) / 1000000.0);
 31
 32     if (includeTimer) {
 33         if (getitimer(ITIMER_REAL, &itv) == -1)
 34             errExit("getitimer");
 35         printf("  %6.2f  %6.2f",
 36                 itv.it_value.tv_sec + itv.it_value.tv_usec / 1000000.0,
 37                 itv.it_interval.tv_sec + itv.it_interval.tv_usec / 1000000.0);
 38     }
 39
 40     printf("\n");
 41     callNum++;
 42 }
 43
 44 static void
 45 sigalrmHandler(int sig)
 46 {
 47     gotAlarm = 1;
 48 }
 49
 50 int
 51 main(int argc, char *argv[])
 52 {
 53     struct itimerval itv;
 54     clock_t prevClock;
 55     int maxSigs;                /* Number of signals to catch before exiting */
 56     int sigCnt;                 /* Number of signals so far caught */
 57     struct sigaction sa;
 58
 59     if (argc > 1 && strcmp(argv[1], "--help") == 0)
 60         usageErr("%s [secs [usecs [int-secs [int-usecs]]]]\n", argv[0]);
 61
 62     sigCnt = 0;
 63
 64     sigemptyset(&sa.sa_mask);
 65     sa.sa_flags = 0;
 66     sa.sa_handler = sigalrmHandler;
 67     if (sigaction(SIGALRM, &sa, NULL) == -1)
 68         errExit("sigaction");
 69
 70     /* Set timer from the command-line arguments */
 71
 72     itv.it_value.tv_sec = (argc > 1) ? getLong(argv[1], 0, "secs") : 2;
 73     itv.it_value.tv_usec = (argc > 2) ? getLong(argv[2], 0, "usecs") : 0;
 74     itv.it_interval.tv_sec = (argc > 3) ? getLong(argv[3], 0, "int-secs") : 0;
 75     itv.it_interval.tv_usec = (argc > 4) ? getLong(argv[4], 0, "int-usecs") : 0;
 76
 77     /* Exit after 3 signals, or on first signal if interval is 0 */
 78
 79     maxSigs = (itv.it_interval.tv_sec == 0 &&
 80                 itv.it_interval.tv_usec == 0) ? 1 : 3;
 81
 82     displayTimes("START:", FALSE);
 83     if (setitimer(ITIMER_REAL, &itv, NULL) == -1)
 84         errExit("setitimer");
 85
 86     prevClock = clock();
 87     sigCnt = 0;
 88
 89     for (;;) {
 90
 91         /* Inner loop consumes at least 0.5 seconds CPU time */
 92
 93         while (((clock() - prevClock) * 10 / CLOCKS_PER_SEC) < 5) {
 94             if (gotAlarm) {                     /* Did we get a signal? */
 95                 gotAlarm = 0;
 96                 displayTimes("ALARM:", TRUE);
 97
 98                 sigCnt++;
 99                 if (sigCnt >= maxSigs) {
100                     printf("That's all folks\n");
101                     exit(EXIT_SUCCESS);
102                 }
103             }
104         }
105
106         prevClock = clock();
107         displayTimes("Main: ", TRUE);
108     }
109 }
